@@ -24,8 +24,9 @@
  *    Zou Nan hai <nanhai.zou@intel.com>
  */
 #include "intel_xvmc.h"
-#include "i830_hwmc.h"
-#include "i810_reg.h"
+#include "intel_hwmc.h"
+#include "i830_reg.h"
+#include "i965_reg.h"
 #include "brw_defines.h"
 #include "brw_structs.h"
 
